#include "avr_f64.h"

// Some strange or inefficient constructions maybe due to
// compiler (vc++, gcc) or frontend restrictions. There seem
// to be some flaws in the vc++ implementation of
// __int64 operations. That was of no further interest to me
// as Intel is not my target processor.


/***********************************************************/
static void f_split(float64_t x, int8_t * f_sign, int16_t * f_ex, 
               uint32_t * f_ufrac, uint32_t * f_lfrac)
/***********************************************************/
{ int64_t * i =&x;
* f_ufrac =(*i>>32)&   0xfffffl;
* f_lfrac = *i     &0xffffffffl;
* f_ex  =  (*i>>52)&((LANGEEINS<<11)-LANGEEINS);
* f_sign = (*i>>63)&LANGEEINS;
}

/***********************************************************/
static void f_combi(float64_t * x, int8_t f_sign, int16_t f_ex,
          uint32_t  f_ufrac,uint32_t  f_lfrac)
/***********************************************************/
{ int64_t * i=x;
  //int64_t   f64;

*i=((uint64_t)f_sign<<63)           |
   ((uint64_t)f_ex<<52  )           |
   ((uint64_t)(f_ufrac&0xfffffl)<<32)|
   ((uint64_t)f_lfrac);


}


/***********************************************************/
float64_t f_sd(float32_t fx)
/***********************************************************/
{ int32_t * i;
  int8_t  f_sign;
  int16_t f_ex; 
  uint32_t  f_ufrac,f_lfrac;
  float64_t f64;
  //float32_t x;

 i=&fx; 
 f_ufrac =(*i    )&0x7fffffl;
 f_lfrac = f_ufrac<<29;
 f_ufrac >>=3;
 f_ex  =  (*i>>23)&0xff;
 f_sign = (*i>>31)&1;

 if(f_ex|f_ufrac|f_lfrac) f_ex = f_ex-0x7f+0x3ff;
 f_combi(&f64,f_sign,f_ex,f_ufrac,f_lfrac);


 return(f64);
}



/***********************************************************/
float32_t f_ds(float64_t fx)
/***********************************************************/
{ 
  int8_t  f_sign;
  int16_t f_ex; 
  uint32_t  f_ufrac,f_lfrac,ui32;
  float32_t f32;
  
 f_split(fx,&f_sign,&f_ex,&f_ufrac,&f_lfrac);

 f_lfrac >>= (31-2);

 if(f_ufrac|f_lfrac|f_ex)
	 f_ex=(f_ex-0x3ff+0x7f)&0xff;
 else
	 f_ex=0;
 
 ui32=((uint32_t)f_sign<<31)|((uint32_t)f_ex<<23)|(f_ufrac<<3)|f_lfrac;
 f32= *((float32_t *)&ui32);
 
 return(f32);
}


/***********************************************************/
static int64_t zze(uint32_t U,uint32_t L)
/***********************************************************/
{
return(((int64_t)(U|(LANGEEINS<<20))<<32)|((int64_t)L));
}

/***********************************************************/
static void ezz(int64_t i64,uint32_t * U,uint32_t *L)
/***********************************************************/
{
*U=(i64>>32);
*U=*U&0xfffffl;
*L=(i64&0xffffffffl);

return;
}

/***********************************************************/
void f_showlong(uint64_t i64)
/***********************************************************/
{uint32_t a,b;

a=i64>>32;
b=(i64&0xffffffffl);
//DebugPrintf("showlong %x %x\n",(int32_t)a,(int32_t)b);
return;
}

/***********************************************************/
static void f_addsub2(float64_t * x, float64_t a, float64_t b,
               uint8_t flagadd, uint8_t * flagexd )
/***********************************************************/
{ // add positive doubles 
  uint8_t asig,bsig,i8;
  uint16_t aex,bex,i16;
  uint32_t aufrac,alfrac,bufrac,blfrac,i32;
  uint64_t ia,ib;
  
  f_split(a,&asig,&aex,&aufrac,&alfrac);
  f_split(b,&bsig,&bex,&bufrac,&blfrac);

  if(!(aex|aufrac|alfrac|bex|bufrac|blfrac)){*x=b;return;}
   
  // order acc. magnitude
  i8=0;
  if(aex>bex) i8=1; else 
  if(aex==bex) 
  if(zze(aufrac,alfrac) > zze(bufrac,blfrac)) i8=1;

  
  * flagexd=0;
  if(i8==0) 
  {  //Exch?
      i16=aex;aex=bex;bex=i16;
      i32=aufrac;aufrac=bufrac;bufrac=i32;
      i32=alfrac;alfrac=blfrac;blfrac=i32;
      i8 =asig;asig=bsig;bsig=i8;
      * flagexd=1;
  } 


  i16=(aex-bex);
  if(i16 < 52) {
      ib=zze(bufrac,blfrac);
      ib |=(LANGEEINS<<52);
      ib >>= i16;
      ia=zze(aufrac,alfrac);
      ia |=(LANGEEINS<<52);
      if(flagadd){ //add
          ia+=ib;
          if(ia>>53){ia>>=1;aex++;}
      }else{       // sub
          ia-=ib;
          i8=0;
          while(!(ia>>52)){
           ia<<=1;
           i8++;
           aex--;
           if((aex==0)||(i8==53)){ // Underflow or Zero
              aex=0;aufrac=0l;alfrac=0l;
              break;
           }
          }
      }
      ezz(ia,&aufrac,&alfrac); 
  }
 
  f_combi(x,0,aex,aufrac,alfrac);
}


/***********************************************************/
static void f_setsign(float64_t * x, int8_t sign)
/***********************************************************/
{ // 1 is neg.
int8_t f_sign;
int16_t f_ex;
uint32_t f_lfrac,f_ufrac;

f_split( *x,&f_sign,&f_ex,&f_ufrac,&f_lfrac);
f_sign=sign;
if(!(f_ex|f_ufrac|f_lfrac))f_sign=0;//unsigned 0
f_combi(x, f_sign, f_ex, f_ufrac,f_lfrac);

}

/***********************************************************/
static uint8_t f_getsign(float64_t x)
/***********************************************************/
{ uint64_t * px=&x;
 return ((uint8_t)(((*px)>>63)&1));
}

/***********************************************************/
float64_t f_add(float64_t a, float64_t b)
/***********************************************************/
{uint8_t signa,signb,signerg;
 uint8_t   flagexd;
 uint64_t  i64;
 float64_t * x = &i64;
 
 signa= f_getsign(a);
 signb= f_getsign(b);
 if(signa^signb)
 {  // diff. signs
   f_addsub2(x,a,b,0,&flagexd); // calc a-b
   signerg= ((flagexd^signa))&1;  
 }else{            // eq.   signs
   f_addsub2(x,a,b,1,&flagexd); // calc a+b
   signerg= signa;
 }
 f_setsign(x,signerg);

return(i64);
}

/***********************************************************/
float64_t f_sub(float64_t a, float64_t b)
/***********************************************************/
{uint8_t signb;
 float64_t  bloc=b;
 uint64_t  i64;
 float64_t * x = &i64;
 
 signb= f_getsign(bloc);
 signb ^= 1;
 f_setsign(&bloc,signb);
 i64=f_add(a,bloc);
 return(i64);

}


/***********************************************************/
static uint64_t f_mult1( uint32_t a, uint32_t b)
/***********************************************************/
{ //multiply 32x32=64
  return((uint64_t)a*(uint64_t)b);
} 

/***********************************************************/
float64_t f_mult(float64_t fa, float64_t fb)
/***********************************************************/
{ //multiply doubles 
  uint8_t  asig,bsig,ui8;
  uint16_t aex,bex;
  int16_t  eex;
  uint32_t a,b,x,y;
  uint64_t ia,ib,ic;
  uint64_t  i64;
  float64_t * fx = &i64;
  
  f_split(fa,&asig,&aex,&a,&b);
  f_split(fb,&bsig,&bex,&x,&y);
  
  if((!(a|b|aex))||
     (!(x|y|bex))){ //0*
      asig=0;aex=0;a=0l;b=0l;
  }else{
	  a |= (LANGEEINS<<20);
	  x |= (LANGEEINS<<20);
	  asig ^= bsig;
	  eex=aex=aex+bex-0x3ff;
      if(eex>2047){ ///Overflow, saturate for now, keep sign
	      aex=2047;a=0xfffffl;b=0xffffffffl;
      }else{
		  if(eex<0){ ///Underflow, set to 0, keep sign
			  aex=0;a=0l;b=0l;
		  }else{               
              ia=f_mult1(b,y);
              ib=f_mult1(a,y)+f_mult1(b,x);
              ic=f_mult1(a,x);
              ia  =((ic<<32)|(ia>>32));   
              ui8 =(uint8_t)(((ia>>1)+(ib>>1))>>63); // Overflow ?
              ia +=ib;
              //f_showlong(ib);
              ic  = ic+(((uint64_t)ui8)<<32); // concider Ovfl
              ui8 = (uint8_t)(ic>>41);         // val. >= 2.0 ?

              ic <<= (51-39);                   // MSB at correct location
              ic  &=  (((uint64_t)(0x1ff))<<(44));
              ic  |=  (ia>>(63-43));

              if(ui8){
                 ic >>=1;
                 aex++;
                 if(aex>2047){aex=2047;a=0xfffffl;b=0xffffffffl;}
              }    
  
              b=(uint32_t)(ic&0xffffffffl);
              a=((uint32_t)(ic>>32))&0xfffff;                 
               }
      }
  }
  f_combi(fx,asig,aex,a,b);
  return(i64);

}

/***********************************************************/
float64_t f_inverse(float64_t fa )
/***********************************************************/
{ // inverse
  // iterate: x=1/a -> x(k+1)=x(k)(2-a*x(k))

  float64_t x,aa,xh,xz,xe,x0;
  //float64_t fa=ffa;
  int16_t  k;
  uint8_t  asig,bsig;
  uint16_t aex,bex;
  uint32_t a,b,uia,uib;
  int32_t  eex;
  uint64_t  i64;
  float64_t * fx = &i64;
 
   
  f_split(fa,&asig,&aex,&a,&b);
  if((!(a|b|aex))){ // divide by 0 wont work
	   asig=0;eex=2047;uia=0xfffffl;uib=0xffffffffl; // saturate for now
  }else{   
	  f_combi(&aa,0,0x3fe,a ,b ); //Convergence radius is 1 ! 
	  f_combi(&xz,0,0x400,0l,0l); // that's 2.0
	  f_combi(&xe,0,0x3ff,0l,0l); // that's 1.0
 
	  x=aa;
	  for(k=0;k<10;k++){
		xh=f_mult(aa,x);

		x0=f_sub (xe,xh);      
        f_split(x0,&bsig,&bex,&uia,&uib);
        if((0x3ff-bex)>52) break; // x*a==1 !
 
		xh=f_sub(xz,xh);
		x=f_mult(x ,xh);
	  }

	  f_split(x,&bsig,&bex,&uia,&uib);
	  eex = -((int32_t)aex-0x3ff)+0x3ff-1+bex-0x3ff;
	  f_split(x,&bsig,&bex,&uia,&uib);
	  //printf("%x\n",bex);
  }
  f_combi(fx,asig,(uint16_t)eex,uia,uib); 
  return(i64);
}

/***********************************************************/
float64_t f_div(float64_t fa, float64_t fb )
/***********************************************************/
{ 
  float64_t x;

  x=f_inverse(fb);
  x=f_mult(fa,x);
  return(x);
}
