//***************************************************************************
//
//***************************************************************************

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "common.h"
#include "main.h"
#include "dprintf.h"

#include "avr_f64.h"

#define IA   ((int32_t)16807)
#define IM   ((int32_t)2147483647)
#define IQ   ((int32_t)127773)
#define IR   ((int32_t)2836)
#define MASK ((int32_t)123459876)


/******************************************************************************/
int32_t rand32() {
/******************************************************************************/
// 32 random Bits, only 31 on AVR, MSB always 0
static int32_t idum=2;
int32_t  k;
      idum ^= MASK;
      k=(idum)/IQ;
      idum=IA*(idum-k*IQ)-IR*k;
      if (idum < 0) idum += IM;
      k=(idum);
      idum ^= MASK;
      return k;
}

/******************************************************************************/
float32_t rand32f() {
/******************************************************************************/
return    ((((float32_t)((uint16_t)(rand32()&0xffffl)))/65535.0)*2.0)-1.0;

}

/******************************************************************************/
float32_t rand_gaussian() {
/******************************************************************************/
// Marsaglia Polar Method
float32_t r,y1,y2;

do{
y1=(2.0*rand32f()-1.0);
y2=(2.0*rand32f()-1.0);
r=y1*y1+y2*y2;
}while(r>1.0);

return (y1*sqrt((-2.0*log(r))/r));

}

/***********************************************************/
void f_check(void)
/***********************************************************/
{ float32_t a,b;
float32_t fa,fb,ferg32,ferr;
int32_t k;
float64_t ferg64,sda,sdb;

#define EPP (1e-6)

  float64_t erg;
  int16_t n;
  float32_t x[]={2.9, 1.999,
                 2.9 , 1.75,
				 3.5 , 1.75 ,
                 1.75  , 3.5,
                 1.75  , 0.0, 
                 0.0  , 8.1,
                 3.5  , 1.75,
                 3.5  , 3.5,
			     3.5,-2.0,
			     3.5,-3.5,
			     3.5,-4.5,
			     -3.5,5.0,
			     -3.5,-8.5,
			     3.5,-2.0,
			     3.5,-2.0,
			     3.5,-2.0,
                  1.0, 1.0,
                    1.0,1.1,
                    1.1,1.1,
                    1.2,1.2,
                    1.5,1.5,
                    1.0/3.0,1.0/7.0,
                    0.0 , 1.0,
                    -1.5,1.5,
                    1.5,-1.5,
                    -1.5,-1.5,
                   //12345678.0,78919.8189173, //diff 10^-3: only 16significant digits
                   1234567.0,7891.8189173,
                   1.0+1.0/177.0, 3.1415926,
                   //3e-38,2e-38
                   3e-12,2e14,
                   0.0,2.0, 
                   0.0,0.5, 
                   0.0,0.0 
                    //1.0/3.0,1.0/6.0
                 
  };
  float32_t xm[]={1.0, 1.0,
                    1.0,1.1,
                    1.1,1.1,
                    1.2,1.2,
                    1.5,1.5,
                    1.0/3.0,1.0/7.0,
                    0.0 , 1.0,
                    -1.5,1.5,
                    1.5,-1.5,
                    -1.5,-1.5,
                   //12345678.0,78919.8189173, //diff 10^-3: only 16significant digits
                   1234567.0,7891.8189173,
                   1.0+1.0/177.0, 3.1415926,
                   //3e-38,2e-38
                   3e-12,2e14,
                   0.0,2.0, 
                   0.0,0.5, 
                   0.0,0.0 
                    //1.0/3.0,1.0/6.0
                 };
 float32_t xi[]={ 0.5,
                  1.5,
                 -1.5,
                  1e3,
                  1e-3
  
                 };

#if(0)  // add/sub checks
   n= sizeof(x)/sizeof(float32_t);

   for(k=0;k<n;k+=2){
    
     a=x[k];b=x[k+1];
	 sda=f_sd(a);sdb=f_sd(b);
     erg=f_add(sda,sdb);
     //erg=sda;

      DebugPrintf("%f \t %f  \t %f \t%f",
	  (float)f_ds(erg),
	  (float)a,
	  (float)b,
	  (float)(a+b-f_ds(erg)));
     if(fabs(a+b-f_ds(erg))>EPP)  DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }

   for(k=0;k<n;k+=2){
     a=x[k];b=x[k+1];
     sda=f_sd(a);sdb=f_sd(b);
     sdb=f_sub(sda,sdb);
     erg=sdb;

      DebugPrintf("%f \t %f  \t %f \t%f",
	  (float)f_ds(erg),
	  (float)a,
	  (float)b,
	  (float)(a-b-f_ds(erg)));
     if(fabs(a-b-f_ds(erg))>EPP)  DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }
 
   DebugPrintf("***********************************\n");
#endif

#if(0)  // multiply checks
   for(k=0;k<51;k++){
     a=1.0;b=1.0+1.0/(1.0*((int64_t)1<<k));
     sda=f_sd(a);sdb=f_sd(b);
     sda=f_mult(sda,sdb);
     erg=sda;

      DebugPrintf("%f \t %f  \t %f \t%f",
	  (float)f_ds(erg),
	  (float)a,
	  (float)b,
	  (float)(a*b-f_ds(erg)));
     if(fabs(a*b-f_ds(erg))>EPP)  DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }

   n= sizeof(xm)/sizeof(float32_t);

   DebugPrintf("*****************************************************\n") ; 
   
   for(k=0;k<n;k+=2){
     a=xm[k];b=xm[k+1];
     sda=f_sd(a);sdb=f_sd(b);
     sda=f_mult(sda,sdb);
     erg=sda;

      DebugPrintf("%f \t %f  \t %f \t%f",
	  (float)f_ds(erg),
	  (float)a,
	  (float)b,
	  (float)(a*b-f_ds(erg)));
     if(fabs(a*b-f_ds(erg))>EPP)  DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }
#endif

#if(0)  // two times check
  
   n= sizeof(x)/sizeof(float32_t);

   DebugPrintf("*****************************************************\n") ; 
   
   for(k=0;k<n;k+=1){
     a=x[k];
     sda=f_sd(a);
     //sdb=f_sd(a);

     //sda=f_mult(sda,sda);
     //sda=f_add(sda,sda);
     //sda=f_div(sda,sda);
     sda=f_inverse(sda);
     erg=sda;

      DebugPrintf("%f \t %f  \t %f \t%f",
	  (float)f_ds(erg),
	  (float)a,
	  (float)a,
	  (float)(1/a-f_ds(erg)));
     if(fabs(1/a-f_ds(erg))>EPP)  DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }
#endif

#if(0)  //inverse checks
   n= sizeof(x)/sizeof(float32_t);
   for(k=0;k<n;k++){
     a=x[k];
     erg=f_inverse(f_sd(a));

      DebugPrintf("%f \t %f  \t %f ",
	  (float)f_ds(erg),
	  (float)a,
	  (float)(a*f_ds(erg)-1.0));
     if(fabs(a*f_ds(erg)-1.0)>EPP) DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }
#endif

#if(0)  // divide checks
   for(k=0;k<51;k++){
     a=1.0;b=1.0+1.0/(1.0*((int64_t)1<<k));
     erg=f_div(f_sd(a),f_sd(b));

     DebugPrintf("%f \t %f  \t %f %f \t",(float)f_ds(erg),(float)a,(float)b,(float)(a/b-f_ds(erg)));
     if(fabs(a/b-f_ds(erg))>EPP) DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }

   n= sizeof(xm)/sizeof(float32_t);
   if(1)
   printf("*****************************************************\n") ; 
   
   for(k=0;k<n;k+=2){
     a=xm[k];b=xm[k+1];
     erg=f_div(f_sd(a),f_sd(b));

     DebugPrintf("%f \t %f  \t %f %f \t",(float)f_ds(erg),(float)a,(float)b,(float)(a/b-f_ds(erg)));
     if(fabs(a/b-f_ds(erg))>EPP) DebugPrintf("******\n") ;else  DebugPrintf("\n"); 
   }
  
#endif

#if(1)  // random checks
k=0;
while(1){
  k++;
  if(!(k&0xff)) DebugPrintf("%d\n",k);

  fa=(float32_t)(100000.0*rand_gaussian());
  fb=(float32_t)(100000.0*rand_gaussian());

  ferg64=f_add(f_sd(fa),f_sd(fb));
  ferg32=f_ds(ferg64);  

  ferr=fabs(fabs((fa+fb)/ferg32-1.0));
  if(ferr>EPP){
	  DebugPrintf("add%.f %f %f %f ",(float)ferr,(float)ferg32,(float)fa,(float)fb); 
	  if(ferr>EPP) DebugPrintf("****\n");else DebugPrintf("\n");
  }

  ferg64=f_sub(f_sd(fa),f_sd(fb));
  ferg32=f_ds(ferg64);  
  ferr=fabs(fabs((fa-fb)/ferg32-1.0));
  if(ferr>EPP){
	  DebugPrintf("sub%f %f %f %f ",(float)ferr,(float)ferg32,(float)fa,(float)fb); 
	  if(ferr>EPP) DebugPrintf("****\n");else DebugPrintf("\n");
  }

  ferg64=f_mult(f_sd(fa),f_sd(fb));
  ferg32=f_ds(ferg64);  
  ferr=fabs(fabs((fa*fb)/ferg32-1.0));
  if(ferr>EPP){
	  DebugPrintf("mul%f %f %f %f ",(float)ferr,(float)ferg32,(float)fa,(float)fb); 
	  if(ferr>EPP) DebugPrintf("****\n");else DebugPrintf("\n");
  }

  ferg64=f_div(f_sd(fa),f_sd(fb));
  ferg32=f_ds(ferg64);  
  ferr=fabs(fabs((fa/fb)/ferg32-1.0));
  if(ferr>EPP){
	  DebugPrintf("div%f %f %f %f ",(float)ferr,(float)ferg32,(float)fa,(float)fb); 
	  if(ferr>EPP) DebugPrintf("****\n");else DebugPrintf("\n");
  }
}
#endif

}


/*****************************************************************************
*   Function name : main
*****************************************************************************/

int main(void)
{ int32_t k,i,m;   
  float32_t r32,r32_a,r32_b,s32,r1,r2,r3,r4,r5,r6,r7,r8;	  
  float64_t r64;	  

	
    Initialization(); 
    DebugPrintf("\n");
 

k=0;
r32=0.0;
for (;;){
 
/*
 for(k=0;k<100;k++)
 r64=f_div(f_sd(1.5),f_sd(7.2));
 DebugPrintf("******\n");
*/
 f_check();  
}

return 0;    
} // ende  main

